from __future__ import division, print_function, absolute_import
from functools import wraps
from lygadgets import pya, isGUI, message, message_loud

from lymask.utilities import lys, LayerSet, insert_layer_tab, gui_view
from lymask.library import dbu, as_region, fast_sized, fast_smoothed, set_threads


all_func_dict = {}
def dpStep(step_fun):
    ''' Each step must accept one argument that is cell, plus optionals, and not return

        steps are added to all_steps *in the order they are defined*
    '''
    all_func_dict[step_fun.__name__] = step_fun
    return step_fun


@dpStep
def check_floorplan(cell, fp_safe=50):
    ''' Checks for floorplan. If you didn't make one, this makes one.
    '''
    if cell.shapes(lys.FLOORPLAN).is_empty():
        message('Warning: No floorplan found. Making one, but you should do this yourself')
        fp_box = cell.dbbox()  # this assumes that, if FLOORPLAN is present, that DRC has verified it forms the extent
        fp_box.enlarge(fp_safe, fp_safe)
        cell.shapes(lys.FLOORPLAN).insert(fp_box)


@dpStep
def flatten(cell):
    cell.flatten(True)


@dpStep
def erase_text_and_other_junk(cell):
    for layname in lys.keys():
        lay = lys[layname]
        for text in cell.shapes(lay).each(pya.Shapes.STexts):
            cell.shapes(lay).erase(text)
        # zero width paths


# Some peole use this, others have already converted
# @dpStep
# def convert_wgs(cell):
    # has_Si = pya.Region()
    # for si in ['wg_deep', 'wg_shallow']:
    #     has_Si.insert(cell.shapes(lys[si]))
    # ^ new version-ish

#     union(cell, 'wg_deep', 'wg_shallow')
#     layer_bool(cell, lys.wg_deep, lys.wg_shallow,
#                pya.EdgeProcessor.ModeBNotA, layC=lys.wg_shallow)
#     cell.clear(lys.wg_deep)
#     cell.move(lys.dp_temp, lys.wg_deep)


# @dpStep
# def nanowire_heal(cell):
#     delta = 0.01 / dbu
#     layer_resize(cell, lys.m2_nw, delta, layC=lys.dp_temp)
#     layer_resize(cell, lys.dp_temp, -delta, layC=lys.m2_nw)


@dpStep
def processor(cell, thread_count=1, remote_host=None):
    if remote_host is not None:
        message_loud('Automatic remote hosting is not yet supported')
    set_threads(thread_count)


@dpStep
def nanowire_sleeve(cell, Delta=2.5, delta=0.2, do_photo=True):
    Delta /= dbu
    delta /= dbu  # new naming convention?
    for dp_lay in ['m2_nw_photo', 'm2_nw_ebeam']:
        cell.clear(lys[dp_lay])
    nw_region = as_region(cell, 'm2_nw')
    nw_compressed = fast_smoothed(nw_region)
    ebeam_region = fast_sized(nw_compressed, Delta + delta) - nw_region
    cell.shapes(lys.m2_nw_ebeam).insert(ebeam_region)
    if do_photo:
        phoas_region = as_region(cell, 'FLOORPLAN') - fast_sized(nw_compressed, Delta - delta)
        cell.shapes(lys.m2_nw_photo).insert(phoas_region)


@dpStep
def waveguide_sleeve(cell, Delta_nw_si=2.0, Delta=2.0, delta=0.2, do_photo=True):
    ''' Does a bulk-sleeve for waveguide full, but first adds it under nanowires.
        Endcaps end where the explicit waveguide ends.
    '''
    Delta_nw_si /= dbu
    Delta /= dbu
    delta /= dbu
    for dp_lay in ['wg_full_photo', 'wg_full_ebeam']:
        cell.clear(lys[dp_lay])

    # add silicon under the nanowires
    nw_compressed = fast_smoothed(as_region(cell, 'm2_nw'))
    wg_explicit = as_region(cell, 'wg_deep')
    nw_except_on_wg = nw_compressed - fast_sized(wg_explicit, Delta_nw_si)
    wg_all = fast_sized(nw_except_on_wg, Delta_nw_si) + wg_explicit

    # do the bulk-sleeve
    # wg_compressed = filter_large_polygons(wg_all)
    wg_compressed = fast_smoothed(wg_all)
    ebeam_region = fast_sized(wg_compressed, Delta + delta) - wg_all
    cell.shapes(lys.wg_full_ebeam).insert(ebeam_region)
    if do_photo:
        phoas_region = as_region(cell, 'FLOORPLAN') - fast_sized(wg_compressed, Delta - delta)
        try:
            phoas_region -= as_region(cell, 'wg_deep_photo')
        except KeyError: pass
        cell.shapes(lys.wg_full_photo).insert(phoas_region)


@dpStep
def ground_plane(cell, Delta_gp=15.0, points_per_circle=100):
    Delta_gp /= dbu
    cell.clear(lys.gp_photo)
    # Accumulate everything that we don't want to cover in metal
    gp_exclusion_things = pya.Region()
    for layname in ['wg_deep', 'wg_shallow', 'm1_nwpad',
                    'm4_ledpad', 'm3_res', 'm5_wiring', 'm2_nw',
                    'GP_KO']:
        try:
            gp_exclusion_things += fast_smoothed(as_region(cell, layname))
        except KeyError: pass
    # Where ground plane is explicitly connected to wires, cut it out of the exclusion region
    gnd_explicit = as_region(cell, 'm5_gnd')
    gp_exclusion_tight = gp_exclusion_things - fast_sized(gnd_explicit, Delta_gp)
    # Inflate the buffer around excluded things and pour
    gp_exclusion_zone = fast_sized(gp_exclusion_tight, Delta_gp)
    gp_region = as_region(cell, 'FLOORPLAN') - gp_exclusion_zone

    # Connect to pads
    gp_region.merge()
    gp_region.round_corners(Delta_gp / 5, Delta_gp / 3, points_per_circle)
    gp_region += as_region(cell, 'm5_gnd')
    gp_region.merge()
    cell.shapes(lys.gp_photo).insert(gp_region)


has_precomped = dict()
@dpStep
def precomp(cell, **kwargs):
    '''
        Arguments are keyed by layer name with the value of bias in microns, so for example

            precomp(TOP, wg_deep=0.05, nw_pad=-0.6)

        The behavior is to overwrite the layer.
        Problem if it runs twice is that it expands too much, so it stores which layers have been biased.
        Currently, it doesn't act on this, assuming that if you have run twice that you hit Undo in between.
    '''
    global has_precomped
    for layer_name, bias_um in kwargs.items():
        # do a check for repeated precomp
        if cell in has_precomped.keys():
            if layer_name in has_precomped[cell]:
                pya.MessageBox.info('Dataprep precomp', 'Warning: precompensating {} twice. '.format(layer_name) + '(in this process)\n'
                                    'If the last precomp was not undone with Ctrl-Z, this will turn out wrong', pya.MessageBox.Ok)
                pass#raise RuntimeError('At this time, precomp cannot be run twice on the same layer.')
            else:
                has_precomped[cell].add(layer_name)
        else:
            has_precomped[cell] = set([layer_name])

        # size the layer in place
        bias = bias_um / dbu
        layer_region = as_region(cell, layer_name)
        layer_region.size(bias)
        cell.clear(lys[layer_name])
        cell.shapes(lys[layer_name]).insert(layer_region)


@dpStep
def mask_map(cell, clear_others=False, **kwargs):
    ''' lyp_file is relative to the yml file. If it is None, the same layer properties will be used.
        kwarg keys are destination layers and values are source layers, which can include "+"
    '''
    mask_layer_index = 101
    # merging and moving to new layers
    for dest_layer, src_expression in kwargs.items():
        new_layinfo = pya.LayerInfo(mask_layer_index, 0, dest_layer)
        lys[dest_layer] = new_layinfo
        cell.layout().layer(new_layinfo)
        components = src_expression.split('+')
        for comp in components:
            cell.copy(lys[comp.strip()], lys[dest_layer])
        mask_layer_index += 1
    if isGUI():
        try:
            lv = gui_view()
            add_tab = True
        except UserWarning:
            # No view is selected. We might be in batch mode
            add_tab = False
        if add_tab:
            if lv.is_transacting():
                lv.commit()
                lv.transaction('Adding mask layers')
            insert_layer_tab(tab_name='Masks')
            for dest_layer in kwargs.keys():
                lay_prop = pya.LayerProperties()
                lay_prop.source_name = lys.get_as_LayerInfo(dest_layer).name
                lay_prop.source_layer = lys.get_as_LayerInfo(dest_layer).layer
                lay_prop.source_datatype = lys.get_as_LayerInfo(dest_layer).datatype
                lv.init_layer_properties(lay_prop)
                lv.insert_layer(lv.end_layers(), lay_prop)
    if clear_others:
        for any_layer in lys.keys():
            if any_layer not in kwargs.keys():
                cell.clear(lys[any_layer])


@dpStep
def align_corners(cell):
    ''' Puts little boxes in the corners so lithography tools all see the same
        Goes through all the layers present in the layout (does not depend on currently loaded layer properties)
    '''
    ly = cell.layout()
    all_layers = ly.layer_infos()

    # get the actual floorplan this time
    if cell.shapes(lys.FLOORPLAN).size() != 1:
        raise RuntimeError('align_corners needs a FLOORPLAN to work consisting of exactly one polygon')
    for fp in cell.shapes(lys.FLOORPLAN).each():
        fp_box = fp.dbbox()

    corner_mark = pya.DBox(0, 0, 1, 1)
    for marked_layer in all_layers:
        if marked_layer.name in ['FLOORPLAN']:  # put exceptions here
            continue
        # do some boolean here to shave off overhangs
        layer_region = as_region(cell, marked_layer)
        layer_region = layer_region & as_region(cell, 'FLOORPLAN')
        cell.clear(lys[marked_layer])
        cell.shapes(lys[marked_layer]).insert(layer_region)

        # put in the markers
        for north_south in (fp_box.top - 1, fp_box.bottom):
            for east_west in (fp_box.right - 1, fp_box.left):
                mark = corner_mark.moved(east_west, north_south)
                if not cell.shapes(ly.layer(marked_layer)).is_empty():
                    cell.shapes(ly.layer(marked_layer)).insert(mark)



'''
    Common functions used by the dataprep steps
'''
from __future__ import division, print_function, absolute_import
from functools import wraps
import pya
from lymask.utilities import active_technology, lys

try:
    dbu = active_technology().dbu
except AttributeError:
    dbu = .001


def as_region(cell, layname):
    ''' Just a convenience brevity function '''
    return pya.Region(cell.shapes(lys[layname]))


def _normal_smoothed(unfiltered_region, deviation=0.1):
    smoothed_region = unfiltered_region.dup()
    smoothed_region.merged_semantics = False
    smoothed_region.smooth(deviation / dbu)
    return smoothed_region


def fast_smoothed(unfiltered_region, deviation=0.1):
    ''' Removes any points that would change the shape by less than this deviation.
        This is used to significantly decrease the number of points prior to sizing

        Note: multicore does not work well with this, but it turns out to be pretty fast
    '''
    # if something goes wrong, you can fall back to regular here by uncommenting
    return _normal_smoothed(unfiltered_region, deviation)

    temp_region = unfiltered_region.dup()
    temp_region.merged_semantics = False

    output_region = pya.Region()
    tp = pya.TilingProcessor()
    tp.input('in1', temp_region)
    tp.output('out1', output_region)
    tp.queue("_output(out1, in1.smoothed({}))".format(deviation / dbu))
    tp.tile_size(2000., 2000.)
    tp.tile_border(5 * deviation, 5 * deviation)
    tp.threads = _thread_count
    tp.execute('Smoothing job')
    return output_region


_thread_count = None
def set_threads(thread_count):
    ''' Set to None to disable parallel processing '''
    global _thread_count
    if thread_count == 1:
        thread_count = None
    _thread_count = thread_count


def fast_sized(input_region, xsize):
    # if something goes wrong, you can fall back to regular here by uncommenting
    if _thread_count is None:
        return input_region.sized(xsize)
    else:
        output_region = pya.Region()
        tp = pya.TilingProcessor()
        tp.input('in1', input_region)
        tp.output('out1', output_region)
        tp.queue("_output(out1, in1.sized({}))".format(xsize))
        tp.tile_size(2000., 2000.)
        tp.tile_border(2 * xsize, 2 * xsize)
        tp.threads = _thread_count
        tp.execute('Sizing job')
        return output_region



from __future__ import division, print_function, absolute_import
from lygadgets import isGUI, pya, lyp_to_layerlist
from lygadgets.technology import Technology, klayout_last_open_technology


#: This global variable to be deprecated
_active_technology = Technology.technology_by_name(klayout_last_open_technology())
def active_technology():
    return _active_technology


def set_active_technology(tech_name):
    if not Technology.has_technology(tech_name):
        raise ValueError('Technology not found. Available are {}'.format(Technology.technology_names()))
    global _active_technology
    _active_technology = Technology.technology_by_name(tech_name)
    reload_lys(tech_name)
# end deprecation


def tech_layer_properties(pya_tech=None):
    ''' Returns the file containing the main layer properties
    '''
    if pya_tech is None:
        pya_tech = active_technology()
    return pya_tech.eff_path(pya_tech.eff_layer_properties_file())


def tech_dataprep_layer_properties(pya_tech=None):
    ''' Returns the file containing the main layer properties
    '''
    if pya_tech is None:
        pya_tech = active_technology()
    return pya_tech.eff_path('dataprep/klayout_layers_dataprep.lyp')


def gui_view():
    patch_environment()  # makes sure the Application attribute gets spoofed into the standalone
    import pya
    lv = pya.Application.instance().main_window().current_view()
    if lv is None:
        raise UserWarning("No view selected. Make sure you have an open layout.")
    return lv


def gui_active_layout():
    ly = gui_view().active_cellview().layout()
    if ly is None:
        raise UserWarning("No layout. Make sure you have an open layout.")
    return ly


def gui_active_cell():
    cell = gui_view().active_cellview().cell
    if cell is None:
        raise UserWarning("No cell. Make sure you have an open layout.")
    return cell


def gui_active_technology():
    pass # todo

class LayerSet(dict):
    ''' getitem returns the logical layer (integer) that can be used in pya functions,
        but you have to set the active_layout first

        It is keyed by the layer name such as 'si_wg',
        but it will succeed if the __getitem__ argument is
            1. integer, in which case, we assume it describes the physical layer to be converted to the logical layer
            2. pya.LayerInfo, in which case, it is ready to be processed by pya

        It also offers object-like attributes that can be gotten, set, and deleted like lys.si_wg.
        If you have a layer like "si_n+", you must use the __getitem__ style

        A note on iterating: this does not work because it side-steps the getitem code::

            for layname, lay in lys.items():

        Instead do this:

            for layname in lys.keys():
                lay = lys[layname]
    '''
    active_layout = None

    def __call__(self, *args, **kwargs):
        return self.__getitem__(args[0])

    def __getattr__(self, attrname):
        if attrname in self.keys():
            return self.__getitem__(attrname)

    def __setattr__(self, attrname, val):
        if attrname in ['active_layout']:  # exceptions
            self.__dict__[attrname] = val
        else:
            self.__setitem__(attrname, val)

    def __delattr__(self, attrname):
        del self[attrname]

    def get_as_LayerInfo(self, item):
        try:
            val = dict.__getitem__(self, item)
        except KeyError as err:
            if type(item) is int:
                val = pya.LayerInfo(item, 0)
            elif type(item) is pya.LayerInfo:
                val = item
            else:
                raise
        return val

    def __getitem__(self, item):
        val = self.get_as_LayerInfo(item)
        try:
            answer = self.active_layout.layer(val)
        except AttributeError as err:
            if "no attribute 'layer'" in err.args[0]:
                err.args = ('You didn\'t set active_layout for this LayerSet, so you can not get items from it', )
            raise
        return answer

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError('Key must be string. Got {}.'.format(type(key)))
        if not isinstance(value, pya.LayerInfo):
            raise TypeError('Value must be pya.LayerInfo. Got {}.'.format(type(value)))
        if not value.is_named():
            value.name = key
        dict.__setitem__(self, key, value)

    @classmethod
    def fromFile(cls, filename):
        new_obj = cls()
        all_layers = lyp_to_layerlist(filename)
        for one_layer in all_layers:
            try:
                group_members = one_layer['group-members']
            except KeyError:  # it is a real layer
                short_name = name2shortName(one_layer['name'])
                new_obj[short_name] = source2pyaLayerInfo(one_layer['source'])
            else:  # it is a group
                if not isinstance(group_members, list):
                    group_members = [group_members]
                for memb in group_members:
                    short_name = name2shortName(memb['name'])
                    new_obj[short_name] = source2pyaLayerInfo(memb['source'])
        return new_obj

    def append(self, other):
        for layname in other.keys():
            if layname in self.keys():
                raise ValueError('Layer is doubly defined: {}'.format(layname))
            self[layname] = other.get_as_LayerInfo(layname)

    def appendFile(self, filename):
        other = LayerSet.fromFile(filename)
        other.active_layout = self.active_layout
        self.append(other)


def name2shortName(name_str):
    ''' Good to have this function separate because
        it may differ for different naming conventions.

        Reassign with::

            lymask.utilities.name2shortName = someOtherFunction
    '''
    if name_str is None:
        raise IOError('This layer has no name')
    components = name_str.split(' - ')
    if len(components) > 1:
        short_name = components[1]
    else:
        short_name = components[0]
    return short_name


def source2pyaLayerInfo(source_str):
    layer_str = source_str.split('@')[0]
    layer, datatype = layer_str.split('/')
    return pya.LayerInfo(int(layer), int(datatype))


def insert_layer_tab(lyp_file=None, tab_name=None):
    ''' Also updates lys, but if any of the layers are already there, it does nothing.
        If lyp_file is None, creates an empty layer list or does nothing if not in GUI mode.
    '''
    if lyp_file is not None and lys is not None:
        try:
            lys.appendFile(lyp_file)
        except ValueError as err:
            if 'doubly defined' in err.args[0]:
                return
            else:
                raise
    if isGUI():
        lv = gui_view()
        i_new_tab = lv.num_layer_lists()
        lv.rename_layer_list(0, 'Designer')
        lv.insert_layer_list(i_new_tab)
        lv.current_layer_list = i_new_tab
        if lyp_file is not None:
            lv.load_layer_props(lyp_file)
        if tab_name is not None:
            lv.rename_layer_list(i_new_tab, tab_name)


lys = LayerSet()
def reload_lys(technology=None):
    global lys
    lys.clear()
    try:
        lyp_file = tech_layer_properties(Technology.technology_by_name(technology))
        lys.appendFile(lyp_file)
    except (FileNotFoundError, AttributeError):
        print('No lyp file found. Likely that technology hasn\'t loaded yet, or you don\'t have the standalone klayout')


reload_lys()




''' This stuff only runs in GUI mode '''
from lygadgets import pya
import glob
import os

from lymask.invocation import gui_main
from lymask.utilities import insert_layer_tab, tech_dataprep_layer_properties

DEFAULT_TECH = 'OLMAC'

def registerMenuItems():
    menu = pya.Application.instance().main_window().menu()
    s1 = "tools_menu.dataprep_menu"
    if not menu.is_menu(s1):
        menu.insert_menu('tools_menu.end', 'dataprep', 'Mask Dataprep')


global item_counter
item_counter = 0
def _gen_new_action(func):
    ''' There a strange bug where pya.Actions get managed to the same location in memory.
        Same with the _Signals that are created when on_triggered is set.

        Assigning them to global variables with different names seems to work.
        It also works if you step through in a debugger.
        It does NOT work if you use locals()

        This function will create functions and action triggers correctly
    '''
    global item_counter
    item_str = 'action_item%s' % item_counter
    func_str = 'action_function%s' % item_counter
    globals()[item_str] = pya.Action()
    globals()[func_str] = func
    globals()[item_str].on_triggered = globals()[func_str]
    item_counter += 1
    return globals()[item_str]


def _gen_dataprep_action(dataprep_file):
    def wrapped():
        gui_main(dataprep_file)
    return _gen_new_action(wrapped)


def dataprep_yml_to_menu(dataprep_file):
    ''' Goes through all .yml files in the given directory and adds a menu item for each one
        These files are passed into the drc-like engine that uses Region to do dataprep steps in python
    '''
    menu = pya.Application.instance().main_window().menu()
    subloop_name = os.path.splitext(os.path.basename(dataprep_file))[0]
    action = _gen_dataprep_action(dataprep_file)
    action.title = 'Run {}.yml'.format(subloop_name)
    if subloop_name == 'default':
        action.shortcut = 'Shift+Ctrl+P'
        menu.insert_separator('tools_menu.dataprep.begin', 'SEP')
        menu.insert_item('tools_menu.dataprep.begin', subloop_name, action)
    else:
        menu.insert_item('tools_menu.dataprep.end', subloop_name, action)


def reload_dataprep_menu(tech_name=None):
    if tech_name is None:
        tech_name = DEFAULT_TECH
    dataprep_dir = pya.Technology.technology_by_name(tech_name).eff_path('dataprep')
    for dataprep_file in glob.iglob(dataprep_dir + '/*.yml'):
        dataprep_yml_to_menu(dataprep_file)
    # Now put in the layers refresh
    menu = pya.Application.instance().main_window().menu()
    layer_action = _gen_new_action(lambda: insert_layer_tab(tech_dataprep_layer_properties(), tab_name='Dataprep'))
    layer_action.title = 'Refresh layer display'
    # layer_action.shortcut = 'Shift+Ctrl+P'
    menu.insert_separator('tools_menu.dataprep.begin', 'SEP2')
    menu.insert_item('tools_menu.dataprep.begin', 'dataprep_layer_refresh', layer_action)



from setuptools import setup
import os


def readme():
    with open('README.md', 'r') as fx:
      return fx.read()


setup(name='lymask',
      version='0.0.6',
      description='Mask dataprep with klayout',
      long_description=readme(),
      author='Alex Tait, Adam McCaughan, Sonia Buckley, Jeff Chiles, Jeff Shainline, Rich Mirin, Sae Woo Nam',
      author_email='alexander.tait@nist.gov',
      license='MIT',
      packages=['lymask'],
      install_requires=['klayout', 'lygadgets>=0.1.14', 'PyYAML'],
      entry_points={'console_scripts': ['lymask=lymask.invocation:cm_main']},
      package_data={'': ['*.lym']},
      include_package_data=True,
      cmdclass={},
      )



'''
    Entry points from GUI, command line, and API
'''
from __future__ import division, print_function, absolute_import
import os
import yaml
import argparse
from lygadgets import pya, message, Technology

from lymask import __version__
from lymask.utilities import gui_view, gui_active_layout, \
                             active_technology, set_active_technology, \
                             tech_dataprep_layer_properties, \
                             lys, insert_layer_tab
from lymask.steps import all_func_dict


parser = argparse.ArgumentParser(description="Command line mask dataprep")
parser.add_argument('infile', type=argparse.FileType('rb'),
                    help='the input gds file')
parser.add_argument('ymlspec', nargs='?', default=None,
                    help='YML file that describes the dataprep steps and parameters. Can be relative to technology')
parser.add_argument('-o', '--outfile', nargs='?', default=None,
                    help='The output file. Default is to tack "_proc" onto the end')
parser.add_argument('-t', '--technology', nargs='?', default=None,
                    help='The name of technology to use. Must be visible in installed technologies')
parser.add_argument('-v', '--version', action='version', version=f'%(prog)s v{__version__}')

def cm_main():
    ''' This one uses the klayout standalone '''
    args = parser.parse_args()
    batch_main(args.infile.name, ymlspec=args.ymlspec, outfile=args.outfile, technology=args.technology)


def _main(layout, ymlfile, tech_obj=None):
    # todo: figure out which technology we will be using and its layer properties
    # todo: reload lys using technology
    with open(ymlfile) as fx:
        step_list = yaml.load(fx)
    insert_layer_tab(tech_dataprep_layer_properties(tech_obj), tab_name='Dataprep')
    for func_info in step_list:
        func = all_func_dict[func_info[0]]
        try:
            kwargs = func_info[1]
        except IndexError:
            kwargs = dict()
        for TOP_ind in layout.each_top_cell():
            # call it
            func(layout.cell(TOP_ind), **kwargs)
    return layout


def gui_main(ymlfile=None):
    layout = gui_active_layout()
    lys.active_layout = layout

    gui_view().transaction('Mask Dataprep')
    try:
        processed = _main(layout, ymlfile=ymlfile, technology=None)  # todo: get the technology from the selection menu
    finally:
        gui_view().commit()


def batch_main(infile, ymlspec=None, technology=None, outfile=None):
    if outfile is None:
        outfile = infile[:-4] + '_proc.gds'
    # Load it
    layout = pya.Layout()
    layout.read(infile)
    lys.active_layout = layout
    # Find the yml file
    if ymlspec is not None and os.path.exists(os.path.realpath(ymlspec)):
        ymlfile = ymlspec
        if technology is not None:
            set_active_technology(technology)
        tech_obj = active_technology()
        if technology is None:
            message('Using the last used technology: {}'.format(tech_obj.name))
    else:
        if technology is None:
            raise ValueError('When specifying a relative dataprep file, you must also provide a technology.')

        tech_obj = Technology.technology_by_name(technology)
        set_active_technology(technology)
        if ymlspec is None:
            # default dataprep test
            ymlfile = tech_obj.eff_path('dataprep/default.yml')
        else:
            # find path to tech
            if not ymlspec.endswith('.yml'):
                ymlspec += '.yml'
            ymlfile = tech_obj.eff_path(os.path.join('dataprep', ymlspec))
    # Process it
    # lys.appendFile(tech_dataprep_layer_properties(tech_obj))
    processed = _main(layout, ymlfile=ymlfile, tech_obj=tech_obj)
    # Write it
    processed.write(outfile)
